@@ -1,12 +1,14 @@
 # Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 # For details: https://github.com/PyCQA/pylint/blob/main/LICENSE
+# Copyright (c) https://github.com/PyCQA/pylint/blob/main/CONTRIBUTORS.txt
+
 from typing import Tuple
 
-__version__ = "2.11.1"
+__version__ = "2.13.3"
 
 
 def get_numversion_from_version(v: str) -> Tuple:
-    """Kept for compatibility reason
+    """Kept for compatibility reason.
 
     See https://github.com/PyCQA/pylint/issues/4399
     https://github.com/PyCQA/pylint/issues/4420,
