@@ -1,29 +1,17 @@
-# -*- coding: utf-8 -*-
-# Copyright (c) 2015-2018 Claudiu Popa <pcmanticore@gmail.com>
-# Copyright (c) 2017 Łukasz Rogalski <rogalski.91@gmail.com>
-# Copyright (c) 2018 ssolanki <sushobhitsolanki@gmail.com>
-# Copyright (c) 2018 Ville Skyttä <ville.skytta@upcloud.com>
-
 # Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
-# For details: https://github.com/PyCQA/pylint/blob/master/COPYING
+# For details: https://github.com/PyCQA/pylint/blob/main/LICENSE
+# Copyright (c) https://github.com/PyCQA/pylint/blob/main/CONTRIBUTORS.txt
+
+"""Visitor doing some postprocessing on the astroid tree.
 
-"""
-Visitor doing some postprocessing on the astroid tree.
 Try to resolve definitions (namespace) dictionary, relationship...
 """
-from __future__ import print_function
-
 import collections
 import os
 import traceback
 
 import astroid
-from astroid import bases
-from astroid import exceptions
-from astroid import manager
-from astroid import modutils
-from astroid import node_classes
-
+from astroid import nodes
 
 from pylint.pyreverse import utils
 
@@ -34,27 +22,27 @@ def _iface_hdlr(_):
 
 
 def _astroid_wrapper(func, modname):
-    print("parsing %s..." % modname)
+    print(f"parsing {modname}...")
     try:
         return func(modname)
-    except exceptions.AstroidBuildingException as exc:
+    except astroid.exceptions.AstroidBuildingException as exc:
         print(exc)
-    except Exception as exc:  # pylint: disable=broad-except
+    except Exception:  # pylint: disable=broad-except
         traceback.print_exc()
+    return None
 
 
 def interfaces(node, herited=True, handler_func=_iface_hdlr):
     """Return an iterator on interfaces implemented by the given class node."""
-    # FIXME: what if __implements__ = (MyIFace, MyParent.__implements__)...
     try:
-        implements = bases.Instance(node).getattr("__implements__")[0]
-    except exceptions.NotFoundError:
+        implements = astroid.bases.Instance(node).getattr("__implements__")[0]
+    except astroid.exceptions.NotFoundError:
         return
-    if not herited and implements.frame() is not node:
+    if not herited and implements.frame(future=True) is not node:
         return
     found = set()
     missing = False
-    for iface in node_classes.unpack_infer(implements):
+    for iface in nodes.unpack_infer(implements):
         if iface is astroid.Uninferable:
             missing = True
             continue
@@ -62,7 +50,7 @@ def interfaces(node, herited=True, handler_func=_iface_hdlr):
             found.add(iface)
             yield iface
     if missing:
-        raise exceptions.InferenceError()
+        raise astroid.exceptions.InferenceError()
 
 
 class IdGeneratorMixIn:
@@ -72,17 +60,44 @@ class IdGeneratorMixIn:
         self.id_count = start_value
 
     def init_counter(self, start_value=0):
-        """init the id counter
-        """
+        """Init the id counter."""
         self.id_count = start_value
 
     def generate_id(self):
-        """generate a new identifier
-        """
+        """Generate a new identifier."""
         self.id_count += 1
         return self.id_count
 
 
+class Project:
+    """A project handle a set of modules / packages."""
+
+    def __init__(self, name=""):
+        self.name = name
+        self.uid = None
+        self.path = None
+        self.modules = []
+        self.locals = {}
+        self.__getitem__ = self.locals.__getitem__
+        self.__iter__ = self.locals.__iter__
+        self.values = self.locals.values
+        self.keys = self.locals.keys
+        self.items = self.locals.items
+
+    def add_module(self, node):
+        self.locals[node.name] = node
+        self.modules.append(node)
+
+    def get_module(self, name):
+        return self.locals[name]
+
+    def get_children(self):
+        return self.modules
+
+    def __repr__(self):
+        return f"<Project {self.name!r} at {id(self)} ({len(self.modules)} modules)>"
+
+
 class Linker(IdGeneratorMixIn, utils.LocalsVisitor):
     """Walk on the project tree and resolve relationships.
 
@@ -116,32 +131,22 @@ class Linker(IdGeneratorMixIn, utils.LocalsVisitor):
         # visited project
         self.project = project
 
-    def visit_project(self, node):
-        """visit a pyreverse.utils.Project node
+    def visit_project(self, node: Project) -> None:
+        """Visit a pyreverse.utils.Project node.
 
-         * optionally tag the node with a unique id
+        * optionally tag the node with a unique id
         """
         if self.tag:
             node.uid = self.generate_id()
         for module in node.modules:
             self.visit(module)
 
-    def visit_package(self, node):
-        """visit an astroid.Package node
+    def visit_module(self, node: nodes.Module) -> None:
+        """Visit an astroid.Module node.
 
-         * optionally tag the node with a unique id
-        """
-        if self.tag:
-            node.uid = self.generate_id()
-        for subelmt in node.values():
-            self.visit(subelmt)
-
-    def visit_module(self, node):
-        """visit an astroid.Module node
-
-         * set the locals_type mapping
-         * set the depends mapping
-         * optionally tag the node with a unique id
+        * set the locals_type mapping
+        * set the depends mapping
+        * optionally tag the node with a unique id
         """
         if hasattr(node, "locals_type"):
             return
@@ -150,12 +155,12 @@ class Linker(IdGeneratorMixIn, utils.LocalsVisitor):
         if self.tag:
             node.uid = self.generate_id()
 
-    def visit_classdef(self, node):
-        """visit an astroid.Class node
+    def visit_classdef(self, node: nodes.ClassDef) -> None:
+        """Visit an astroid.Class node.
 
-         * set the locals_type and instance_attrs_type mappings
-         * set the implements list and build it
-         * optionally tag the node with a unique id
+        * set the locals_type and instance_attrs_type mappings
+        * set the implements list and build it
+        * optionally tag the node with a unique id
         """
         if hasattr(node, "locals_type"):
             return
@@ -171,18 +176,19 @@ class Linker(IdGeneratorMixIn, utils.LocalsVisitor):
         node.instance_attrs_type = collections.defaultdict(list)
         for assignattrs in node.instance_attrs.values():
             for assignattr in assignattrs:
-                self.handle_assignattr_type(assignattr, node)
+                if not isinstance(assignattr, nodes.Unknown):
+                    self.handle_assignattr_type(assignattr, node)
         # resolve implemented interface
         try:
             node.implements = list(interfaces(node, self.inherited_interfaces))
         except astroid.InferenceError:
-            node.implements = ()
+            node.implements = []
 
-    def visit_functiondef(self, node):
-        """visit an astroid.Function node
+    def visit_functiondef(self, node: nodes.FunctionDef) -> None:
+        """Visit an astroid.Function node.
 
-         * set the locals_type mapping
-         * optionally tag the node with a unique id
+        * set the locals_type mapping
+        * optionally tag the node with a unique id
         """
         if hasattr(node, "locals_type"):
             return
@@ -195,8 +201,8 @@ class Linker(IdGeneratorMixIn, utils.LocalsVisitor):
     link_class = visit_classdef
     link_function = visit_functiondef
 
-    def visit_assignname(self, node):
-        """visit an astroid.AssignName node
+    def visit_assignname(self, node: nodes.AssignName) -> None:
+        """Visit an astroid.AssignName node.
 
         handle locals_type
         """
@@ -205,94 +211,86 @@ class Linker(IdGeneratorMixIn, utils.LocalsVisitor):
         if hasattr(node, "_handled"):
             return
         node._handled = True
-        if node.name in node.frame():
-            frame = node.frame()
+        if node.name in node.frame(future=True):
+            frame = node.frame(future=True)
         else:
             # the name has been defined as 'global' in the frame and belongs
             # there.
             frame = node.root()
-        try:
-            if not hasattr(frame, "locals_type"):
-                # If the frame doesn't have a locals_type yet,
-                # it means it wasn't yet visited. Visit it now
-                # to add what's missing from it.
-                if isinstance(frame, astroid.ClassDef):
-                    self.visit_classdef(frame)
-                elif isinstance(frame, astroid.FunctionDef):
-                    self.visit_functiondef(frame)
-                else:
-                    self.visit_module(frame)
+        if not hasattr(frame, "locals_type"):
+            # If the frame doesn't have a locals_type yet,
+            # it means it wasn't yet visited. Visit it now
+            # to add what's missing from it.
+            if isinstance(frame, nodes.ClassDef):
+                self.visit_classdef(frame)
+            elif isinstance(frame, nodes.FunctionDef):
+                self.visit_functiondef(frame)
+            else:
+                self.visit_module(frame)
 
-            current = frame.locals_type[node.name]
-            values = set(node.infer())
-            frame.locals_type[node.name] = list(set(current) | values)
-        except astroid.InferenceError:
-            pass
+        current = frame.locals_type[node.name]
+        frame.locals_type[node.name] = list(set(current) | utils.infer_node(node))
 
     @staticmethod
     def handle_assignattr_type(node, parent):
-        """handle an astroid.assignattr node
+        """Handle an astroid.assignattr node.
 
         handle instance_attrs_type
         """
-        try:
-            values = set(node.infer())
-            current = set(parent.instance_attrs_type[node.attrname])
-            parent.instance_attrs_type[node.attrname] = list(current | values)
-        except astroid.InferenceError:
-            pass
+        current = set(parent.instance_attrs_type[node.attrname])
+        parent.instance_attrs_type[node.attrname] = list(
+            current | utils.infer_node(node)
+        )
 
-    def visit_import(self, node):
-        """visit an astroid.Import node
+    def visit_import(self, node: nodes.Import) -> None:
+        """Visit an astroid.Import node.
 
         resolve module dependencies
         """
         context_file = node.root().file
         for name in node.names:
-            relative = modutils.is_relative(name[0], context_file)
+            relative = astroid.modutils.is_relative(name[0], context_file)
             self._imported_module(node, name[0], relative)
 
-    def visit_importfrom(self, node):
-        """visit an astroid.ImportFrom node
+    def visit_importfrom(self, node: nodes.ImportFrom) -> None:
+        """Visit an astroid.ImportFrom node.
 
         resolve module dependencies
         """
         basename = node.modname
         context_file = node.root().file
         if context_file is not None:
-            relative = modutils.is_relative(basename, context_file)
+            relative = astroid.modutils.is_relative(basename, context_file)
         else:
             relative = False
         for name in node.names:
             if name[0] == "*":
                 continue
             # analyze dependencies
-            fullname = "%s.%s" % (basename, name[0])
+            fullname = f"{basename}.{name[0]}"
             if fullname.find(".") > -1:
                 try:
-                    # TODO: don't use get_module_part,
-                    # missing package precedence
-                    fullname = modutils.get_module_part(fullname, context_file)
+                    fullname = astroid.modutils.get_module_part(fullname, context_file)
                 except ImportError:
                     continue
             if fullname != basename:
                 self._imported_module(node, fullname, relative)
 
     def compute_module(self, context_name, mod_path):
-        """return true if the module should be added to dependencies"""
+        """Return true if the module should be added to dependencies."""
         package_dir = os.path.dirname(self.project.path)
         if context_name == mod_path:
             return 0
-        if modutils.is_standard_module(mod_path, (package_dir,)):
+        if astroid.modutils.is_standard_module(mod_path, (package_dir,)):
             return 1
         return 0
 
     def _imported_module(self, node, mod_path, relative):
-        """Notify an imported module, used to analyze dependencies"""
+        """Notify an imported module, used to analyze dependencies."""
         module = node.root()
         context_name = module.name
         if relative:
-            mod_path = "%s.%s" % (".".join(context_name.split(".")[:-1]), mod_path)
+            mod_path = f"{'.'.join(context_name.split('.')[:-1])}.{mod_path}"
         if self.compute_module(context_name, mod_path):
             # handle dependencies
             if not hasattr(module, "depends"):
@@ -302,48 +300,16 @@ class Linker(IdGeneratorMixIn, utils.LocalsVisitor):
                 mod_paths.append(mod_path)
 
 
-class Project:
-    """a project handle a set of modules / packages"""
-
-    def __init__(self, name=""):
-        self.name = name
-        self.path = None
-        self.modules = []
-        self.locals = {}
-        self.__getitem__ = self.locals.__getitem__
-        self.__iter__ = self.locals.__iter__
-        self.values = self.locals.values
-        self.keys = self.locals.keys
-        self.items = self.locals.items
-
-    def add_module(self, node):
-        self.locals[node.name] = node
-        self.modules.append(node)
-
-    def get_module(self, name):
-        return self.locals[name]
-
-    def get_children(self):
-        return self.modules
-
-    def __repr__(self):
-        return "<Project %r at %s (%s modules)>" % (
-            self.name,
-            id(self),
-            len(self.modules),
-        )
-
-
 def project_from_files(
     files, func_wrapper=_astroid_wrapper, project_name="no name", black_list=("CVS",)
 ):
-    """return a Project from a list of files or modules"""
+    """Return a Project from a list of files or modules."""
     # build the project representation
-    astroid_manager = manager.AstroidManager()
+    astroid_manager = astroid.manager.AstroidManager()
     project = Project(project_name)
     for something in files:
         if not os.path.exists(something):
-            fpath = modutils.file_from_modpath(something.split("."))
+            fpath = astroid.modutils.file_from_modpath(something.split("."))
         elif os.path.isdir(something):
             fpath = os.path.join(something, "__init__.py")
         else:
@@ -351,14 +317,13 @@ def project_from_files(
         ast = func_wrapper(astroid_manager.ast_from_file, fpath)
         if ast is None:
             continue
-        # XXX why is first file defining the project.path ?
         project.path = project.path or ast.file
         project.add_module(ast)
         base_name = ast.name
         # recurse in package except if __init__ was explicitly given
         if ast.package and something.find("__init__") == -1:
             # recurse on others packages / modules if this is a package
-            for fpath in modutils.get_module_files(
+            for fpath in astroid.modutils.get_module_files(
                 os.path.dirname(ast.file), black_list
             ):
                 ast = func_wrapper(astroid_manager.ast_from_file, fpath)
